###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

###############################################################################
# This file provides the ProgressBar class - a simple wrapper around tqdm
# progress bars. It is used in download/upload methods and provides the
# benefit of remembering state information and printing only when in
# verbose mode.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from coscine.client import Client
from tqdm import tqdm

###############################################################################

class ProgressBar:

	def __init__(self, client: Client, filesize: int, filename: str, \
								mode: str, callafter: function = None):

		MODES = {
			"UP": "↑",
			"DOWN": "↓"
		}
		if mode not in MODES:
			raise ValueError(mode)

		self.client = client
		if client.verbose:
			self.bar = tqdm(total = filesize, unit = "B", unit_scale = True,
									desc = "%s %s" % (MODES[mode], filename))
		self.callafter = callafter
		self.n = 0

###############################################################################

	def update(self, chunksize: int):
		self.n += chunksize
		if self.client.verbose:
			self.bar.update(chunksize)
		if self.callafter:
			self.callafter(chunksize)

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from .client import Client
	from .resource import Resource
	from .MetadataForm import MetadataForm
import os
import json
from .progress import ProgressBar

###############################################################################

class Object:

###############################################################################

	client: Client
	resource: Resource
	data: dict
	metadata: dict
	name: str
	size: int
	type: str

###############################################################################

	CHUNK_SIZE = 4096

###############################################################################

	def __init__(self, resource: Resource, data: dict, metadata: dict):
		self.client = resource.client
		self.resource = resource
		self.data = data
		self.metadata = metadata
		self.name = data["Name"]
		self.size = data["Size"]
		self.type = data["Kind"]

###############################################################################

	def __str__(self) -> str:
		return json.dumps(self.data, indent = 4)

###############################################################################

	def content(self) -> bytes:
		uri = self.client.uri("Blob", "Blob", self.resource.id, self.name)
		data = self.client.get(uri).content
		return data

###############################################################################

	def download(self, path = "./", callback = None):
		# uri = self.data["Action"]["Download"] # Not working, Duh!
		uri = self.client.uri("Blob", "Blob", self.resource.id, self.name)
		response = self.client.get(uri, stream = True)
		path = os.path.join(path, self.name)
		fd = open(path, "wb")
		bar = ProgressBar(self.client, self.size, self.name, "DOWN", callback)
		for chunk in response.iter_content(chunk_size = self.CHUNK_SIZE):
			fd.write(chunk)
			bar.update(len(chunk))
		fd.close()

###############################################################################

	def delete(self):
		uri = self.client.uri("Blob", "Blob", self.resource.id, self.name)
		self.client.delete(uri)

###############################################################################

	def update(self, metadata: MetadataForm):
		if type(metadata) is MetadataForm:
			metadata = metadata.generate()
		uri = self.client.uri("Tree", "Tree", self.resource.id, self.name)
		self.client.put(uri, data = metadata)

###############################################################################

	def MetadataForm(self) -> MetadataForm:
		form = self.resource.MetadataForm()
		form.parse(self.metadata)
		return form

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

###############################################################################
# This file contains the main class of the coscine python package.
# The client class is the backbone and manager of everything contained
# within the coscine python module.
###############################################################################

import json
import urllib
import requests
from requests.exceptions import RequestException
import colorama
from requests.models import Response
from .banner import BANNER
from .about import __version__
from .exceptions import *
from .project import Project
from .static import StaticServer
from .ProjectForm import ProjectForm

###############################################################################

class Client:

#######################################
# Class member variable declaration
#######################################

	lang: str
	verbose: bool
	session: requests.Session
	loglevel: str
	version: str
	static: StaticServer

#######################################
# Class Constructor
###############################################################################

	def __init__(self, token: str, lang: str = "en", verbose: bool = True, \
					loglevel: list[str] = ["LOG", "INFO", "WARN", "REQUEST", "DATA"]):

		LANG = ("en", "de")
		if type(token) is not str:
			raise TypeError("Argument 'token' must be of type string!")
		if lang not in LANG:
			raise ValueError("Invalid value in argument 'lang'!")

		self.session = requests.Session()
		self.session.headers = {
			"Authorization": "Bearer " + token,
			"User-Agent": "Coscine Python Client %s" % __version__
		}
		self.lang = lang
		self.verbose = verbose
		self.loglevel = loglevel
		self.version = __version__
		self.static = StaticServer(self)
		if verbose:
			colorama.init(autoreset=True)
			print(BANNER)

###############################################################################

	def log(self, msg: str, level: str = "LOG"):
		LEVELS = {
			"LOG": colorama.Fore.MAGENTA,
			"INFO": colorama.Fore.GREEN,
			"WARN": colorama.Fore.YELLOW,
			"REQUEST": colorama.Fore.LIGHTGREEN_EX,
			"DATA": colorama.Fore.LIGHTBLUE_EX
		}

		if level not in LEVELS:
			raise ValueError(level)

		if not level in self.loglevel:
			return

		if self.verbose:
			print(LEVELS[level] + "[%s] " % level + msg)

###############################################################################

	def is_outdated(self) -> bool:
		uri = "https://pypi.org/pypi/coscine/json"
		data = requests.get(uri).json()
		version = data["info"]["version"]
		if version != __version__:
			msg = "Using Coscine Client version %s but latest version is %s.\n"\
					"Consider updating the package by running:\n" \
					"py -m pip install --user --upgrade coscine\n" \
					% (__version__, version)
			self.log(msg)
			return True
		else:
			return False

###############################################################################

	@staticmethod
	def uri(api: str, endpoint: str, *args) -> str:
		BASE = "https://coscine.rwth-aachen.de/coscine/api/Coscine.Api.%s/%s"
		ENDPOINTS = (
			"Blob",
			"Metadata",
			"Organization",
			"Project",
			"Resources",
			"Tree",
			"User",
			"ActivatedFeatures"
		)

		if api not in ENDPOINTS:
			raise ValueError("Argument 'api' does not specify a valid endpoint!")

		uri = BASE % (api, endpoint)
		for arg in args:
			if arg is None:
				continue
			uri += "/" + urllib.parse.quote(arg, safe="")
		return uri

###############################################################################

	def _request(self, method: str, uri: str, **kwargs) -> Response:
		if "data" in kwargs and type(kwargs["data"]) is dict:
			kwargs["headers"] = {"Content-Type": "application/json;charset=utf-8"}
			kwargs["data"] = json.dumps(kwargs["data"])

		try:
			self.log("%s %s" % (method, uri), "REQUEST")
			if "data" in kwargs:
				try:
					self.log(json.dumps(json.loads(kwargs["data"]), indent=4), "DATA")
				except :
					pass
			response = self.session.request(method, uri, **kwargs)
			response.raise_for_status()
			if response.content:
				try:
					self.log(json.dumps(response.json(), indent=4), "DATA")
				except:
					pass
			return response
		except requests.exceptions.ConnectionError:
			raise ConnectionError()
		except requests.exceptions.RequestException as e:
			if e.response.status_code == 401:
				raise UnauthorizedError("Invalid API token")
			else:
				raise RequestException()

###############################################################################

	def get(self, uri: str, **kwargs) -> requests.Response:
		return self._request("GET", uri, **kwargs)

###############################################################################

	def put(self, uri: str, **kwargs) -> requests.Response:
		return self._request("PUT", uri, **kwargs)

###############################################################################

	def post(self, uri: str, **kwargs) -> requests.Response:
		return self._request("POST", uri, **kwargs)

###############################################################################

	def delete(self, uri: str, **kwargs) -> requests.Response:
		return self._request("DELETE", uri, **kwargs)

###############################################################################

	def projects(self, toplevel: bool = True, **kwargs) -> list[Project]:
		ENDPOINTS = ("Project", "Project/-/topLevel")
		uri = self.uri("Project", ENDPOINTS[toplevel])
		projects = self.get(uri).json()
		filter = []
		for data in projects:
			match = True
			for key, value in kwargs.items():
				if data[key] != value:
					match = False
					break
			if match:
				filter.append(Project(self, data))
		return filter

###############################################################################

	def project(self, displayName: str = None, **kwargs) -> Project:
		if displayName:
			kwargs["displayName"] = displayName
		projects = self.projects(toplevel = True, **kwargs)
		if len(projects) == 1:
			return projects[0]
		elif len(projects) == 0:
			return None
		else:
			raise CoscineException("Undistinguishable projects!")

###############################################################################

	def ProjectForm(self, parent: Project = None) -> ProjectForm:
		vocabulary = {
			"Discipline": self.static.disciplines(),
			"Organization": self.static.organizations(),
			"Visibility": self.static.visibility(),
			"Features": self.static.features()
		}

		form = ProjectForm(self.lang, vocabulary, parent)
		return form

###############################################################################

	def create_project(self, form: ProjectForm) -> Project:
		if type(form) is ProjectForm:
			form = form.generate()

		uri = self.uri("Project", "Project")
		return Project(self, self.post(uri, data = form).json())

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

###############################################################################
# The purpose of the StaticServer class defined in this file is to provide
# static data which is not subject to change. Initially a request is made
# to the Coscine REST API to query such data. Upon further requests a
# cached response is used to speed things up and reduce internet traffic.
# Dynamic data such as the metadata of a project cannot be cached as
# it may be invalidated by further requests or interactions with a
# different client such as the web interface.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from coscine.client import Client

import urllib
import os

###############################################################################

class StaticServer:

	client: Client
	cache: dict

	def __init__(self, client):
		self.client = client
		self.cache = {}

###############################################################################

	def disciplines(self) -> dict:
		LMAP = {
			"de": "displayNameDe",
			"en": "displayNameEn"
		}
		lang = LMAP[self.client.lang]

		if "disciplines" in self.cache:
			data = self.cache["disciplines"]
		else:
			uri = self.client.uri("Project", "Discipline")
			data = self.client.get(uri).json()
			self.cache["disciplines"] = data

		disciplines = {}
		for entry in data:
			disciplines[entry[lang]] = entry
		return disciplines

###############################################################################

	def organizations(self) -> dict:
		organizations = {}
		if "organizations" in self.cache:
			data = self.cache["organizations"]
		else:
			uri = self.client.uri("Organization", "Organization")
			data = self.client.get(uri).json()
		for entry in data["data"]:
			organizations[entry["displayName"]] = entry
		return organizations

###############################################################################

	def visibility(self) -> dict:
		visibility = {}
		uri = self.client.uri("Project", "Visibility")
		data = self.client.get(uri).json()
		for entry in data:
			visibility[entry["displayName"]] = entry
		return visibility

###############################################################################

	def features(self) -> dict:
		features = {}
		uri = self.client.uri("ActivatedFeatures", "ActivatedFeatures")
		data = self.client.get(uri).json()
		for entry in data:
			features[entry[self.client.lang]] = entry
		return features

###############################################################################

	def licenses(self) -> dict:
		licenses = {}
		uri = self.client.uri("Project", "License")
		data = self.client.get(uri).json()
		for entry in data:
			licenses[entry["displayName"]] = entry
		return licenses

###############################################################################

	def resource_types(self) -> dict:
		uri = self.client.uri("Resources", "ResourceType", "types")
		data = self.client.get(uri).json()
		types = {}
		for it in data:
			if it["isEnabled"]:
				types[it["displayName"]] = it
		return types

###############################################################################

	def application_profiles(self) -> dict:
		profiles = {}
		uri = self.client.uri("Metadata", "Metadata", "profiles")
		data = self.client.get(uri).json()
		for entry in data:
			name = urllib.parse.urlparse(entry)[2]
			name = os.path.relpath(name, "/coscine/ap/")
			name = name.upper()
			profiles[name] = entry
		return profiles

###############################################################################

	@staticmethod
	def _parse_application_profile(profile: dict) -> dict:

		def _get_lang(entry, lang):
			for it in entry:
				if it["@language"] == lang:
					return it
			return None

		"""
		Parses an application profile into a human readable python dict
		"""

		W3PREFIX = "http://www.w3.org/ns/shacl#%s"
		data = {}
		profile = profile[0]
		data["id"] = profile["@id"]
		graph = []
		for entry in profile["@graph"]:
			obj = {}
			if W3PREFIX % "name" not in entry:
				continue
			obj["id"] = entry["@id"]
			obj["path"] = entry[W3PREFIX % "path"][0]["@id"]
			obj["order"] = int(entry[W3PREFIX % "order"][0]["@value"])
			if W3PREFIX % "minCount" in entry:
				obj["minCount"] = int(entry[W3PREFIX % "minCount"][0]["@value"])
			if W3PREFIX % "maxCount" in entry:
				obj["maxCount"] = int(entry[W3PREFIX % "maxCount"][0]["@value"])
			obj["name"] = {
				"de": _get_lang(entry[W3PREFIX % "name"], "de")["@value"],
				"en": _get_lang(entry[W3PREFIX % "name"], "en")["@value"]
			}
			if W3PREFIX % "datatype" in entry:
				obj["datatype"] = entry[W3PREFIX % "datatype"][0]["@id"]
				obj["type"] = "literal"
			if W3PREFIX % "class" in entry:
				obj["class"] = entry[W3PREFIX % "class"][0]["@id"]
				obj["datatype"] = obj["class"]
				obj["type"] = "uri"
			graph.append(obj)
		data["graph"] = graph
		return data

###############################################################################

	def application_profile(self, path: str, id: str = None, parse: bool = False) -> dict:
		uri = self.client.uri("Metadata", "Metadata", "profiles", path, id)
		profile = self.client.get(uri).json()
		if parse:
			profile = self._parse_application_profile(profile)
		return profile

###############################################################################


###############################################################################
# Coscine Python3 Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/wiki/-/tree/master/
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# This python wrapper implements a client for the Coscine API.
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.rwth-aachen.de for more information.
###############################################################################

from enum import IntFlag

class FormFlags(IntFlag):
	NONE = 0x0
	REQUIRED = 0x1
	CONTROLLED = 0x2
	SET = 0x4
	FIXED = 0x8


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .client import Client

import os
import json
from .exceptions import *
from .resource import Resource
from .ResourceForm import ResourceForm
from .ProjectForm import ProjectForm
from .MetadataPresetForm import MetadataPresetForm
from .member import Member

###############################################################################

class Project:

#######################################
# Class member variable declaration
#######################################

	client: Client
	data: dict
	id: str
	name: str
	description: str
	principleInvestigators: str
	startDate: str
	endDate: str
	discipline: list[str]
	organization: list[str]

#######################################

	def __init__(self, client: Client, data: dict):
		self.client = client
		self.data = data
		self.id = data["id"]
		self.name = data["displayName"]
		self.description = data["description"]
		self.principleInvestigators = data["principleInvestigators"]
		self.startDate = data["startDate"]
		self.endDate = data["endDate"]
		self.discipline = []
		self.organization = []

###############################################################################

	def __str__(self) -> str:
		return json.dumps(self.data, indent=4)

###############################################################################

	def subprojects(self, **kwargs) -> list:
		uri = self.client.uri("Project", "SubProject", self.id)
		projects = self.client.get(uri).json()
		filter = []
		for data in projects:
			match = True
			for key, value in kwargs.items():
				if data[key] != value:
					match = False
					break
			if match:
				filter.append(Project(self.client, data))
		return filter

###############################################################################

	def resources(self, **kwargs) -> list[Resource]:
		uri = self.client.uri("Project", "Project", self.id, "resources")
		resources = self.client.get(uri).json()
		filter = []
		for data in resources:
			match = True
			for key, value in kwargs.items():
				if data[key] != value:
					match = False
					break
			if match:
				filter.append(Resource(self, data))
		return filter

###############################################################################

	def resource(self, displayName: str = None, **kwargs) -> Resource:
		if displayName:
			kwargs["displayName"] = displayName
		resources = self.resources(**kwargs)
		if len(resources) == 1:
			return resources[0]
		else:
			raise CoscineException("Undistinguishable resources!")

###############################################################################

	def download(self, path: str = "./"):
		path = os.path.join(path, self.displayName)
		if not os.path.isdir(path):
			os.mkdir(path)
		for resource in self.resources():
			resource.download(path = path)

###############################################################################

	def delete(self):
		uri = self.client.uri("Project", "Project", self.id)
		self.client.delete(uri)

###############################################################################

	def members(self) -> list[Member]:
		uri = self.client.uri("Project", "ProjectRole", self.id)
		data = self.client.get(uri).json()
		members = [Member(self, m) for m in data]
		return members

###############################################################################

	def invite(self, email: str, role: str = "Member"):
		ROLES = {
			"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
			"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
		}

		if role not in ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "Project", "invitation")
		data = {
			"projectId": self.data["id"],
			"role": ROLES[role],
			"email": email
		}

		try:
			self.client.log("Inviting [%s] as [%s] to project [%s]." % 
												(email, role, self.id))
			self.client.post(uri, data = data)
		except ServerError:
			self.client.log("User [%s] has invite pending." % email)

###############################################################################

	def _instance(self, link: str) -> dict:
		uri = self.client.uri("Metadata", "Metadata", "instances", self.id, link)
		return self.client.get(uri).json()

###############################################################################

	def ResourceForm(self) -> ResourceForm:
		vocabulary = {
			"type": self.client.static.resource_types(),
			"applicationProfile": self.client.static.application_profiles(),
			"License": self.client.static.licenses(),
			"Visibility": self.client.static.visibility(),
			"Disciplines": self.client.static.disciplines()
		}
		return ResourceForm(self.client.lang, vocabulary)

###############################################################################

	def create_resource(self, resourceForm: ResourceForm, \
					metadataPreset: MetadataPresetForm = None) -> Resource:
		if type(resourceForm) is ResourceForm:
			resourceForm = resourceForm.generate()
		if metadataPreset:
			if type(metadataPreset) is MetadataPresetForm:
				metadataPreset = metadataPreset.generate()
			resourceForm["fixedValues"] = metadataPreset
		uri = self.client.uri("Resources", "Resource", "Project", self.id)
		return Resource(self, self.client.post(uri, data = resourceForm).json())

###############################################################################

	def form(self) -> ProjectForm:
		form = self.client.ProjectForm()
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form: ProjectForm):
		if type(form) is ProjectForm:
			form = form.generate()
		uri = self.client.uri("Project", "Project", self.id)
		self.client.post(uri, data = form)

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

#
# The MetadataPreset form allows for the configuration of default
# values in an application profile.
#


from .form import Form

###############################################################################

class MetadataPresetForm(Form):
	
	def __init__(self, profile: dict, lang: str, entries: list, vocabulary: dict):
		super().__init__("Metadata Preset", lang, entries, vocabulary)
		self.profile = profile

###############################################################################

	def enable(self, key: str, enable: bool = True):
		self._keys[key]["enabled"] = enable

###############################################################################

	def lock(self, key: str, lock: bool = True):
		self._keys[key]["locked"] = lock

###############################################################################

	def generate(self) -> dict:
		metadata = {}

		# Set metadata fields
		for key in self._keys:
			field = self._keys[key]
			path = field["path"]

			if key not in self.store:
				value = None
				metadata[path] = {
					"https://purl.org/coscine/defaultValue": [],
					"https://purl.org/coscine/invisible": [{
						"type": "literal",
						"value": "0"
					}]
				}
				continue
			else:
				value = self.store[key]
			if self.is_controlled(key):
				voc = self.get_vocabulary(key)
				value = voc[value]
			metadata[path] = {
				"https://purl.org/coscine/defaultValue": [{
					"type": field["type"],
					"value": value
				}],
				"https://purl.org/coscine/invisible": [{
					"type": "literal",
					"value": "0"
				}]
			}

		return metadata

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .client import Client
	from .project import Project

###############################################################################

class Member:

# Member variables (only for type hinting, not static!)
###############################################################################

	id: str
	role: str
	data: dict
	client: Client
	project: Project

###############################################################################

	def __init__(self, project: Project, data: dict):
		self.client = project.client
		self.project = project
		self.data = data
		self.name = data["user"]["displayName"]
		self.email = data["user"]["emailAddress"]
		self.id = data["user"]["id"]
		self.role = data["role"]["displayName"]

###############################################################################

	def set_role(self, role: str):
		ROLES = {
			"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
			"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
		}

		if role not in ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "ProjectRole")
		self.data["role"]["id"] = ROLES[role]
		self.data["role"]["displayName"] = role
		self.client.post(uri, data = self.data)

###############################################################################

	def remove(self):
		uri = self.client.uri("Project", "ProjectRole", "project", \
			self.project.id, "user", self.id, "role", self.data["role"]["id"])
		self.client.delete(uri)

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from .exceptions import *
from .form import Form

###############################################################################

KEYS = [{
		"name": {
			"de": "Ressourcentyp",
			"en": "Resource Type"
		},
		"flags": Form.REQUIRED | Form.CONTROLLED,
		"field": "type"
	},{ # This is REQUIRED for rds and s3, but not for Linked Data!
		"name": {
			"de": "Ressourcengröße",
			"en": "Resource Size"
		},
		"flags": Form.SPECIAL,
		"field": "resourceTypeOption"
	},{
		"name": {
			"de": "Ressourcenname",
			"en": "Resource Name"
		},
		"flags": Form.REQUIRED,
		"field": "ResourceName"
	},{
		"name": {
			"de": "Anzeigename",
			"en": "Display Name"
		},
		"flags": Form.REQUIRED,
		"field": "DisplayName"
	},{
		"name": {
			"de": "Ressourcenbeschreibung",
			"en": "Resource Description"
		},
		"flags": Form.REQUIRED,
		"field": "Description"
	},{
		"name": {
			"de": "Disziplin",
			"en": "Discipline"
		},
		"flags": Form.REQUIRED | Form.CONTROLLED | Form.LIST,
		"field": "Disciplines"
	},{
		"name": {
			"de": "Ressourcenschlagwörter",
			"en": "Resource Keywords"
		},
		"flags": Form.NONE,
		"field": "Keywords"
	},{
		"name": {
			"de": "Lizenz",
			"en": "License"
		},
		"flags": Form.CONTROLLED,
		"field": "License"
	},{
		"name": {
			"de": "Verwendungsrechte",
			"en": "Usage Rights"
		},
		"flags": Form.NONE,
		"field": "UsageRights"
	},{
		"name": {
			"de": "Applikationsprofile",
			"en": "Application Profile"
		},
		"flags": Form.REQUIRED | Form.CONTROLLED,
		"field": "applicationProfile"
}]

MAP = {
	"description": "Description",
	"displayName": "DisplayName",
	"resourceName": "ResourceName",
	"keywords": "Keywords",
	"disciplines": "Disciplines",
	"license": "License",
	"resourceTypeOption": "resourceTypeOption",
	"applicationProfile": "applicationProfile",
	"type": "type",
	"usageRights": "UsageRights"
}

###############################################################################

class ResourceForm(Form):

	"""
	Coscine Input Form for creating and editing resources
	"""

###############################################################################

	def __init__(self, lang: str, vocabulary: dict):
		super().__init__("Resource Form", lang, KEYS, vocabulary)

###############################################################################

	def parse(self, data: dict):

		def _parse_disciplines(self, data):
			LSTR = {"en": "displayNameEn", "de": "displayNameDe"}[self._lang]
			disciplines = []
			for discipline in data:
				disciplines.append(discipline[LSTR])
			return disciplines

		for key in data:
			if key in MAP and data[key] != "":
				entry = self.entry(MAP[key])
				if entry is None:
					continue
				name = entry["name"][self._lang]
				flags = entry["flags"]
				value = data[key]
				if flags & Form.CONTROLLED or flags & Form.LIST or flags & Form.SPECIAL:
					if key == "disciplines":
						self[name] = _parse_disciplines(self, value)
					elif key == "resourceTypeOption":
						self[name] = value["Size"]
					elif key == "applicationProfile":
						values = list(self._vocabulary["applicationProfile"].values())
						keys = list(self._vocabulary["applicationProfile"].keys())
						position = values.index(value)
						self[name] = keys[position]
					elif key == "visibility" or key == "license" or key == "type":
						self[name] = value["displayName"]
				else:
					self[name] = value

###############################################################################

	def generate(self) -> dict:

		"""
		Generates JSON-LD formatted representation of resource data

		Raises
		-------
		RequirementError
			When one or more required fields have not been set
		
		Returns
		--------
		JSON-LD formatted resource data
		"""

		def final_value(key, value):
			if self.is_controlled(key):
				return self._vocabulary[field][value]
			else:
				return value

		data = {}
		missing = []
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				value = self.store[key]
				if entry["flags"] & ResourceForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
					data[field] = []
					for v in value:
						data[field].append(final_value(key, v))
				else:
					data[field] = final_value(key, value)

		if missing:
			raise RequirementError(missing)

		size = {}
		if "resourceTypeOption" in data:
			size = {
				"Size": data["resourceTypeOption"]
			}
		elif data["type"]["displayName"] in ("rds", "rds-s3"):
			raise RequirementError("rds or rds-s3 require a size parameter!")
		data["resourceTypeOption"] = size

		data["Visibility"] = {
			"displayName": "Project Members",
			"id": "8ab9c883-eb0d-4402-aaad-2e4007badce6"
		}

		return data

###############################################################################


###############################################################################
# Coscine Python3 Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/wiki/-/tree/master/
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# This python wrapper implements a client for the Coscine API.
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.rwth-aachen.de for more information.
###############################################################################

import os
import json
import urllib
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from requests_toolbelt.multipart.encoder import coerce_data
from tqdm import tqdm
import colorama
from .exceptions import *
from .MetadataForm import MetadataForm
from .ProjectForm import ProjectForm
from .ResourceForm import ResourceForm

###############################################################################

__version__ = "0.4.1"

###############################################################################
# Coscine Banner
###############################################################################

BANNER = colorama.Fore.BLUE + \
"""
                     _             
                    (_)            
   ___ ___  ___  ___ _ _ __   ___  
  / __/ _ \/ __|/ __| | '_ \ / _ \ 
 | (_| (_) \__ \ (__| | | | |  __/ 
  \___\___/|___/\___|_|_| |_|\___| 

""" + colorama.Fore.WHITE + \
"____________________________________\n" + \
"\n   Coscine API Client" + colorama.Fore.YELLOW + " " + __version__ + "\n" + \
colorama.Fore.WHITE + "____________________________________\n\n"

###############################################################################

class CoscineClient:

	"""
	Provides a python client for the Coscine REST API
	"""

###############################################################################

	def __init__(self, token, verbose=True, lang="en"):

		"""
		Constructor of the CoscineClient class

		Parameters
		----------
		token : str
			Your Coscine API token
		verbose : bool
			Enable/Disable command line output
		lang : str
			Set form and controlled vocabulary language
		"""

		if lang != "en" and lang != "de":
			raise ValueError("Expected \"en\" or \"de\" for parameter lang!")

		if type(token) is not str:
			raise ValueError("Expected Coscine API token as string!")

		self.session = requests.Session()
		self.session.headers = {
			"Authorization": "Bearer " + token,
			"User-Agent": "Coscine Python Client v0.3.2"
		}
		self.verbose = verbose
		self.lang = lang
		self.cache = {}
		if verbose:
			colorama.init(autoreset=True)
			self.log(BANNER)
			self.check_version()

###############################################################################

	def check_version(self):

		"""
		Checks for the latest version hosted at pypi and compares it with
		the current version.
		"""

		uri = "https://pypi.org/pypi/coscine/json"
		data = requests.get(uri).json()
		version = data["info"]["version"]
		if version != __version__:
			msg = "Using CoscineClient version %s but latest version is %s.\n"\
					"Consider updating the package by running:\n" \
					"py -m pip install --user --upgrade coscine\n" \
					"This version is OUTDATED and UNSUPPORTED!\n" \
					% (__version__, version)
			self.log(msg, level = "WARN")

###############################################################################

	def __del__(self):
		pass

###############################################################################

	@staticmethod
	def uri(api, endpoint, *args):

		"""
		Coscine API URL builder

		Parameters
		----------
		api : str
			Select the Coscine API (e.g. "Project" or "Metadata")
		endpoint : str
			Select the Coscine API endpoint
		args : vargs
			Include values and additional parameters
		"""

		BASE = "https://coscine.rwth-aachen.de/coscine/api/Coscine.Api.%s/%s"
		uri = BASE % (api, endpoint)
		if len(args) > 0:
			for arg in args:
				uri += "/" + urllib.parse.quote(arg, safe="")
		return uri

###############################################################################

	def log(self, msg, level = None):

		"""
		If logging is enabled you can use this method to print messages
		to the command line interface
		"""

		LEVELS = {
			"INFO": colorama.Fore.GREEN,
			"LOG": colorama.Fore.MAGENTA,
			"WARN": colorama.Fore.YELLOW
		}

		if self.verbose:
			if level and level in LEVELS:
				msg = LEVELS[level] + "[%s] " % level + msg
			print(msg)

###############################################################################

	def _req(self, method, uri, cache = False, **kwargs):

		"""
		HTTP Request handler with simple cache
		"""

		# Some rudimentary caching to speed things up
		if cache and uri in self.cache and method == "GET":
			return self.cache[uri]

		try:
			self.log("%s %s" % (method, uri), level = "LOG")
			response = self.session.request(method, uri, **kwargs)
			response.raise_for_status()
			if cache and method == "GET":
				self.cache[uri] = response
			return response
		except requests.exceptions.ConnectionError as e:
			raise ConnectionError()
		except requests.exceptions.RequestException as e:
			raise ServerError()

###############################################################################

	def get(self, uri, cache = False, **kwargs):
		return self._req("GET", uri, cache = cache, **kwargs)

###############################################################################

	def put(self, uri, **kwargs):
		return self._req("PUT", uri, **kwargs)

###############################################################################

	def post(self, uri, **kwargs):
		return self._req("POST", uri, **kwargs)

###############################################################################

	def delete(self, uri, **kwargs):
		return self._req("DELETE", uri, **kwargs)

###############################################################################

	def get_projects(self, **kwargs):

		"""
		Get a list of projects optionally matching certain criteria
		
		Parameters
		----------
		kwargs : unfolded dictionary containing filter criteria
			May contain any top-level key present in the project json
		
		Returns
		-------
		A list of projects as json dicts or an empty list
		"""

		uri = self.uri("Project", "Project")
		projects = self.get(uri, cache=True).json()
		if kwargs:
			projects = self._filter(projects, kwargs)
		return projects

###############################################################################

	def get_project(self, displayName):

		"""
		Get a specific project by its display name

		Parameters
		----------
		displayName : str
			The display name of the project
		"""

		projects = self.get_projects(displayName = displayName)
		if projects:
			if len(projects) > 1:
				self.log("Projects share common displayName!", level="WARN")
			return projects[0]
		else:
			raise NotFound(displayName)

###############################################################################

	def delete_project(self, project):

		"""
		Delete a project in Coscine
		
		Parameters
		----------
		project : JSON-object containing project information
		"""

		self.log("Deleting project [%s]" % project["projectName"], level="INFO")
		uri = self.uri("Project", "Project", project["id"])
		self.delete(uri)

###############################################################################

	def create_project(self, form):

		"""
		Create a Project in Coscine
		
		Parameters
		----------
		form : ProjectForm
		
		Returns
		-------
		A handle to the new project
		"""

		if type(form) is ProjectForm:
			form = form.generate()
		if type(form) is dict:
			form = json.dumps(form)
		uri = self.uri("Project", "Project")
		return self.post(uri, data=form).json()

###############################################################################

	def download_project(self, project, path="./"):

		"""
		Download a Coscine project including all of its resources to
		a directory on your harddrive
		
		Parameters
		----------
		project : Coscine Project Handle
		path : (optional) Path to the local save directory
		
		Returns
		-------
		Nothing
		"""

		self.log("Downloading project [%s]" % project["projectName"], level="INFO")
		path = os.path.join(path, project["displayName"])
		if not os.path.isdir(path):
			os.mkdir(path)
		for resource in self.get_resources(project):
			self.download_resource(resource, path=path)

###############################################################################

	def get_resources(self, project, **kwargs):

		"""
		Parameters
		----------
		
		Returns
		-------
		
		"""

		uri = self.uri("Project", "Project", project["id"], "resources")
		resources = self.get(uri, cache = True).json()
		if kwargs:
			resources = self._filter(resources, kwargs)
		return resources

###############################################################################

	def get_resource(self, project, displayName):
		resources = self.get_resources(project, displayName = displayName)
		if resources and len(resources) == 1:
			return resources[0]
		else:
			raise NotFound(displayName)

###############################################################################

	def delete_resource(self, resource):

		"""
		Deletes a Resource

		Parameters
		----------
		resource : Coscine Resource Handle

		Returns
		-------
		Nothing
		"""

		self.log("Deleting resource [%s]." % resource["resourceName"], level="INFO")
		uri = self.uri("Resources", "Resource", resource["id"])
		self.delete(uri)

###############################################################################

	def create_resource(self, project, form):

		"""
		Parameters
		----------
		
		Returns
		-------
	
		"""

		if type(form) is ResourceForm:
			form = form.generate()
		if type(form) is dict:
			form = json.dumps(form)
		uri = self.uri("Resources", "Resource", "Project", project["id"])
		headers = {
			"Content-Type": "application/json;charset=utf-8"
		}
		self.post(uri, data=form, headers=headers)

###############################################################################

	def download_resource(self, resource, path="./"):

		"""
		Downloads a resource and all of its files

		Parameters
		----------
		resource : Coscine Resource Handle
			The Resource you want to download

		path : str
			The path to the save directory

		Returns
		-------
		Nothing
		
		"""

		self.log("Downloading resource [%s]" % resource["resourceName"], level="INFO")
		dirpath = os.path.join(path, resource["displayName"])
		if not os.path.isdir(dirpath):
			os.mkdir(dirpath)
		files = self.get_files(resource)
		for file in files:
			filepath = os.path.join(dirpath, file["Path"].strip("/"))
			self.download_file(resource, file["Name"], path=filepath)

###############################################################################

	def get_quota(self, resource):

		"""
		Get the quota used by a Coscine resource in bytes

		Parameters
		----------
		resource : Coscine resource handle

		Returns
		-------
		The used quota of the resource as int
		"""

		uri = self.uri("Blob", "Blob", resource["id"], "quota")
		data = self.get(uri).json()
		quota = int(data["data"]["usedSizeByte"])
		return quota

###############################################################################

	def get_files(self, resource, **kwargs):

		"""
		Get a list of files and their metadata

		Parameters
		----------
		resource : Coscine Resource handle
		kwargs : Optional keyword arguments

		Returns
		-------
		A list of files and their metadata
		"""

		uri = self.uri("Tree", "Tree", resource["id"])
		data = self.get(uri, cache=True).json()
		files = []
		for entry in data["data"]["fileStorage"]:
			info = {
				"Name": entry["Name"],
				"Path": entry["Path"],
				"Size": entry["Size"],
				"Type": entry["Kind"],
				"Provider": entry["Provider"]
			}
			files.append(info)

		if kwargs:
			files = self._filter(files, kwargs)

		return files

###############################################################################

	def get_metadata(self, resource, filename):

		"""
		Get the metadata of a file

		Parameters
		----------
		resource : Coscine Resource Handle
		filename : str

		Returns
		-------
		Metadata of file referenced by filename
		"""

		uri = self.uri("Tree", "Tree", resource["id"], filename)
		data = self.get(uri).json()
		metadata = data["data"]["metadataStorage"][0]
		for key in metadata:
			return metadata[key]

###############################################################################

	def put_metadata(self, resource, filename, metadata):

		"""
		Upload or update metadata for a file

		Parameters
		----------
		resource : Coscine Resource Handle
		filename : str
		metadata : MetadataForm or dict or str

		Returns
		-------
		"""

		if type(metadata) is MetadataForm:
			metadata = metadata.generate()
		if type(metadata) is dict:
			metadata = json.dumps(metadata)
		uri = self.uri("Tree", "Tree", resource["id"], filename)
		self.put(uri, data = metadata)

###############################################################################

	def download_file(self, resource, filename, path):

		"""
		Downloads a file from a resource

		Parameters
		----------
		resource : Coscine Resource Handle
		filename : str
		path : str

		Returns
		-------

		"""

		uri = self.uri("Blob", "Blob", resource["id"], filename)
		response = self.get(uri, stream = True)
		fd = open(path, "wb")
		CHUNK_SIZE = 4096
		if self.verbose:
			filesize = self.get_files(resource, Name = filename)[0]["Size"]
			bar = tqdm(total = filesize, unit = "B", unit_scale = True, \
											desc = "↓ %s" % filename)
		for chunk in response.iter_content(chunk_size = CHUNK_SIZE):
			fd.write(chunk)
			if self.verbose:
				bar.update(len(chunk))
		if self.verbose:
			bar.close()
		fd.close()

###############################################################################

	def upload_file(self, resource, filename, path, metadata):

		"""
		Uploads a file to a resource
		"""

		fd = open(path, "rb")
		self.upload_data(resource, filename, fd, metadata)
		fd.close()

###############################################################################

	def _progress_callback(self, monitor, bar):
		if self.verbose:
			bar.update(monitor.bytes_read - bar.n)

	def upload_data(self, resource, filename, data, metadata):

		"""
		Uploads data to a resource

		Parameters
		----------
		"""

		uri = self.uri("Blob", "Blob", resource["id"], filename)
		self.put_metadata(resource, filename, metadata)
		fields = {
			"files": (filename, data, "application/octet-stream")
		}
		encoder = MultipartEncoder(fields = fields)
		filesize = encoder.len
		if self.verbose:
			bar = tqdm(total = filesize, unit = "B", unit_scale = True, \
				desc = "↑ %s" % filename)
		else:
			bar = None
		monitor = MultipartEncoderMonitor(encoder, \
			lambda callback: self._progress_callback(callback, bar))
		headers = {
			"Content-Type": monitor.content_type
		}
		self.put(uri, data = monitor, headers = headers)
		if self.verbose:
			bar.close()

###############################################################################

	def delete_file(self, resource, filename):

		"""
		Deletes a file from a resource

		Parameters
		----------
		resource : Coscine Resource Handle
		filename : str
		"""

		self.log("Deleting file [%s/%s]" % (resource["resourceName"], filename), level="INFO")
		uri = self.uri("Blob", "Blob", resource["id"], filename)
		try:
			self.delete(uri)
		except ServerError:
			raise CoscineException("File [%s] does not exist" % filename)

###############################################################################

	@staticmethod
	def _get_lang(entry, lang):
		for it in entry:
			if it["@language"] == lang:
				return it
		return None

	@staticmethod
	def _parse_application_profile(profile):
		W3PREFIX = "http://www.w3.org/ns/shacl#%s"
		data = {}
		profile = profile[0]
		data["id"] = profile["@id"]
		graph = []
		for entry in profile["@graph"]:
			obj = {}
			if W3PREFIX % "name" not in entry:
				continue
			obj["id"] = entry["@id"]
			obj["path"] = entry[W3PREFIX % "path"][0]["@id"]
			obj["order"] = int(entry[W3PREFIX % "order"][0]["@value"])
			if W3PREFIX % "minCount" in entry:
				obj["minCount"] = int(entry[W3PREFIX % "minCount"][0]["@value"])
			if W3PREFIX % "maxCount" in entry:
				obj["maxCount"] = int(entry[W3PREFIX % "maxCount"][0]["@value"])
			obj["name"] = {
				"de": CoscineClient._get_lang(entry[W3PREFIX % "name"], "de")["@value"],
				"en": CoscineClient._get_lang(entry[W3PREFIX % "name"], "en")["@value"]
			}
			if W3PREFIX % "datatype" in entry:
				obj["datatype"] = entry[W3PREFIX % "datatype"][0]["@id"]
				obj["type"] = "literal"
			if W3PREFIX % "class" in entry:
				obj["class"] = entry[W3PREFIX % "class"][0]["@id"]
				obj["datatype"] = obj["class"]
				obj["type"] = "uri"
			graph.append(obj)
		data["graph"] = graph
		return data

	def get_application_profile(self, resource, parse=False):

		"""
		Returns the application profile used by a resource

		Parameters
		----------
		resource : Coscine Resource Handle
		parse : bool
			If set to true, converts the application profile to a simplified
			representation
		"""

		uri = self.uri("Metadata", "Metadata", "profiles", \
			resource["applicationProfile"], resource["id"])
		profile = self.get(uri).json()
		if parse:
			profile = CoscineClient._parse_application_profile(profile)
		return profile

###############################################################################

	def MetadataForm(self, project, resource, data=None):

		"""
		Generate a Coscine Metadata Input Form

		Parameters
		----------
		
		Returns
		-------
		A Metadata Input Form
		"""

		form = MetadataForm(self, project, resource)
		if data:
			for key in data:
				form[key] = data[key]
		return form

###############################################################################

	def ProjectForm(self, parent=None, data=None):

		"""
		Parameters
		----------
		
		Returns
		-------
		
		"""

		form = ProjectForm(self, parent=parent)
		if data:
			for key in data:
				form[key] = data[key]
		return form

###############################################################################

	def ResourceForm(self, data=None):

		"""
		Parameters
		----------
		
		Returns
		-------
		
		"""

		form = ResourceForm(self)
		if data:
			for key in data:
				form[key] = data[key]
		return form

###############################################################################

	def get_disciplines(self):

		"""
		"""

		LMAP = {
			"de": "displayNameDe",
			"en": "displayNameEn"
		}
		lang = LMAP[self.lang]
		disciplines = {}
		uri = self.uri("Project", "Discipline")
		data = self.get(uri, cache=True).json()
		for entry in data:
			disciplines[entry[lang]] = entry
		return disciplines

###############################################################################

	def get_organizations(self, filter=None):

		"""
		Parameters
		----------
		
		Returns
		-------
		
		"""

		organizations = {}
		uri = self.uri("Organization", "Organization")
		data = self.get(uri, cache=True).json()
		for entry in data["data"]:
			organizations[entry["displayName"]] = entry
		return organizations

###############################################################################

	def get_visibility(self):
		visibility = {}
		uri = self.uri("Project", "Visibility")
		data = self.get(uri, cache=True).json()
		for entry in data:
			visibility[entry["displayName"]] = entry
		return visibility

###############################################################################

	def get_features(self):
		features = {}
		uri = self.uri("ActivatedFeatures", "ActivatedFeatures")
		data = self.get(uri, cache=True).json()
		for entry in data:
			features[entry[self.lang]] = entry
		return features

###############################################################################

	def get_instance(self, project, link):
		uri = self.uri("Metadata", "Metadata", "instances", \
										project["id"], link)
		instance = self.get(uri, cache=True).json()
		return instance

###############################################################################

	def get_resource_types(self):
		uri = self.uri("Resources", "ResourceType", "types")
		data = self.get(uri, cache=True).json()
		types = {}
		for it in data:
			if it["isEnabled"]:
				types[it["displayName"]] = it
		return types

###############################################################################

	def get_profiles(self):
		profiles = {}
		uri = self.uri("Metadata", "Metadata", "profiles")
		data = self.get(uri, cache=True).json()
		for entry in data:
			name = urllib.parse.urlparse(entry)[2]
			name = os.path.relpath(name, "/coscine/ap/")
			name = name.upper()
			profiles[name] = entry
		return profiles

###############################################################################

	def get_licenses(self):
		licenses = {}
		uri = self.uri("Project", "License")
		data = self.get(uri, cache=True).json()
		for entry in data:
			licenses[entry["displayName"]] = entry
		return licenses

###############################################################################

	def activate_feature(self, project, feature, flag):
		FLAGS = {
			True: "activateFeature",
			False: "deactivateFeature"
		}
		uri = self.uri("ActivatedFeatures", "ActivatedFeatures", \
			project["id"], FLAGS[flag], feature["id"])
		self.get(uri)

###############################################################################

	@staticmethod
	def _filter(data, kwargs):

		"""
		Filters an array of dictionaries according to kwargs
		"""

		filtered = []
		for entry in data:
			for kw in kwargs:
				if kw in entry and entry[kw] == kwargs[kw]:
					filtered.append(entry)
		return filtered



###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from .client import Client
from .project import Project
from .resource import Resource
from .object import Object
from .exceptions import *

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from .client import Client
	from .project import Project
import os
import json
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from .exceptions import *
from .object import Object
from .MetadataForm import MetadataForm
from .MetadataPresetForm import MetadataPresetForm
from .ResourceForm import ResourceForm
from .progress import ProgressBar

###############################################################################

class Resource:

	client: Client
	project: Project
	data: dict
	id: str
	name: str

	def __init__(self, project, data):
		self.client = project.client
		self.project = project
		self.data = data
		self.id = data["id"]
		self.name = data["displayName"]

###############################################################################

	def __str__(self) -> str:
		return json.dumps(self.data, indent=4)

###############################################################################

	def upload(self, key: str, file, metadata: dict = None, callback: function = None):
		if hasattr(file, "read"):
			fd = file
			filename = "MEM"
		elif type(file) is str:
			fd = open(file, "rb")
			filename = file
		else:
			raise CoscineException()

		if metadata:
			if type(metadata) is MetadataForm:
				metadata = metadata.generate()
			uri = self.client.uri("Tree", "Tree", self.id, key)
			self.client.put(uri, data = metadata)

		uri = self.client.uri("Blob", "Blob", self.id, key)
		fields = {"files": (key, fd, "application/octect-stream")}
		encoder = MultipartEncoder(fields = fields)
		bar = ProgressBar(self.client, encoder.len, filename, "UP", callback)
		monitor = MultipartEncoderMonitor(encoder, callback = lambda monitor: \
										bar.update(monitor.bytes_read - bar.n))
		headers = {"Content-Type": monitor.content_type}
		self.client.put(uri, data = monitor, headers = headers)

###############################################################################

	def download(self, path = "./"):
		path = os.path.join(path, self.name)
		if not os.path.isdir(path):
			os.mkdir(path)
		for file in self.objects():
			file.download(path)

###############################################################################

	def delete(self):
		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.delete(uri)

###############################################################################

	def quota(self) -> int:
		uri = self.client.uri("Blob", "Blob", self.id, "quota")
		data = self.client.get(uri).json()
		return int(data("data")["usedSizeByte"])

###############################################################################

	def objects(self, **kwargs) -> list[Object]:
		uri = self.client.uri("Tree", "Tree", self.id)
		data = self.client.get(uri).json()
		fileStorage = data["data"]["fileStorage"]
		metadataStorage = data["data"]["metadataStorage"]
		objects = zip(fileStorage, metadataStorage)
		filter = []
		for f in objects:
			match = True
			for key, value in kwargs.items():
				if f[0][key] != value:
					match = False
					break
			if match:
				filter.append(Object(self, f[0], f[1]))
		return filter

###############################################################################

	def object(self, displayName: str = None) -> Object:
		objects = self.objects()
		for it in objects:
			if it.name == displayName:
				return it
		return None

###############################################################################

	def applicationProfile(self, parse = False):
		return self.client.static.application_profile(
				self.data["applicationProfile"], self.id, parse)

###############################################################################

	def MetadataForm(self, data = None):
		entries = []
		vocabulary = {}
		lang = self.client.lang
		profile = self.applicationProfile(parse = True)
		for element in profile["graph"]:
			flags = MetadataForm.NONE
			if "minCount" in element and element["minCount"] > 0:
				flags |= MetadataForm.REQUIRED
			if "maxCount" in element and element["maxCount"] > 1:
				flags |= MetadataForm.LIST

			if "class" in element:
				flags |= MetadataForm.CONTROLLED
				uri = element["class"]
				instance = self.project._instance(uri)
				if lang not in instance or len(instance[lang]) == 0:
					lang = "en"
				voc = {}
				for rule in instance[lang]:
					voc[rule["name"]] = rule["value"]
				vocabulary[element["name"][lang]] = voc

			entry = {
				"name": element["name"],
				"path": element["path"],
				"field": element["name"][lang],
				"flags": flags,
				"order": element["order"],
				"datatype": element["datatype"],
				"type": element["type"]
			}
			entries.append(entry)

		# Sort the keys according to their application profile order
		entries = sorted(entries, key = lambda x: x["order"])

		form = MetadataForm(profile, self.client.lang, entries, vocabulary)
		if data:
			for key in data:
				form[key] = data[key]
		return form

###############################################################################

	def MetadataPresetForm(self) -> MetadataPresetForm:
		entries = []
		vocabulary = {}
		lang = self.client.lang
		applicationProfile = self.applicationProfile(parse = True)
		for element in applicationProfile["graph"]:
			flags = MetadataPresetForm.NONE
			if "minCount" in element and element["minCount"] > 0:
				flags |= MetadataPresetForm.REQUIRED
			if "maxCount" in element and element["maxCount"] > 1:
				flags |= MetadataPresetForm.LIST

			if "class" in element:
				flags |= MetadataPresetForm.CONTROLLED
				uri = element["class"]
				instance = self.project._instance(uri)
				if lang not in instance or len(instance[lang]) == 0:
					lang = "en"
				voc = {}
				for rule in instance[lang]:
					voc[rule["name"]] = rule["value"]
				vocabulary[element["name"][lang]] = voc
			
			entry = {
				"name": element["name"],
				"path": element["id"],
				"flags": flags,
				"field": element["name"][lang],
				"order": element["order"],
				"datatype": element["datatype"],
				"type": element["type"],
				"locked": False,
				"enabled": True
			}
			entries.append(entry)

		# Sort the keys according to their application profile order
		entries = sorted(entries, key = lambda x: x["order"])
		return MetadataPresetForm(applicationProfile, lang, entries, vocabulary)

###############################################################################

	def form(self) -> ResourceForm:
		form = self.project.ResourceForm()
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form):
		if type(form) is ResourceForm:
			form = form.generate()
		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.post(uri, data = form)

###############################################################################

	def update_preset(self, form):
		data = self.form().generate()
		data["fixedValues"] = form.generate()
		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.post(uri, data = data)

###############################################################################


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

class CoscineException(Exception):
	pass

class ConnectionError(CoscineException):
	pass

class ServerError(CoscineException):
	pass

class VocabularyError(CoscineException):
	pass

class RequirementError(CoscineException):
	pass

class UnauthorizedError(CoscineException):
	pass


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from .exceptions import *
from .form import Form

###############################################################################

class MetadataForm(Form):

	def __init__(self, profile: dict, lang: str, entries: list, vocabulary: dict):
		super().__init__("Metadata Form", lang, entries, vocabulary)
		self.profile = profile

###############################################################################

	def parse(self, data: dict):

		"""
		Parses JSON-LD metadata into a Metadata Input Form
		"""

		k = list(data.keys())[0]
		for path in data[k]:
			for entry in self.profile["graph"]:
				if path == entry["path"]:
					key = entry["name"][self._lang]
					value = data[k][path][0]["value"]
					if self.is_controlled(key):
						voc = self.get_vocabulary(key)
						keys = list(voc.keys())
						values = list(voc.values())
						index = values.index(value)
						value = keys[index]
					self.store[key] = value
					break

###############################################################################

	def generate(self) -> dict:
		metadata = {}
		RDFTYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

		# Set application profile type used by metadata
		metadata[RDFTYPE] = [{
			"type": "uri",
			"value": self.profile["id"]
		}]

		# Collect missing required fields
		missing = []

		# Set metadata fields
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				path = entry["path"]
				value = self.store[key]
				if entry["flags"] & MetadataForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
				if self.is_controlled(key):
					voc = self.get_vocabulary(key)
					value = voc[value]
				metadata[path] = [{
					"value": value,
					"datatype": entry["datatype"],
					"type": entry["type"]
				}]

		# Check missing field list
		if len(missing) > 0:
			raise RequirementError(missing)

		return metadata

###############################################################################



import setuptools
from src.coscine.about import __version__, __author__

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

setuptools.setup(
	name = "coscine",
	version = __version__,
	description = "Coscine Python Client",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	author = __author__,
	author_email = "coscine@itc.rwth-aachen.de",
	license = "MIT License",
	packages = setuptools.find_packages(where="src"),
	keywords = [
		"Coscine", "RWTH Aachen", "Research Data Management"
	],
	install_requires = [
		"requests",
		"requests-toolbelt",
		"tqdm",
		"colorama",
		"boto3"
	],
	url = "https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client",
	project_urls = {
		"Issues":  "https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client/-/issues",
		"Wiki": "https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client/-/wikis/home"
	},
	classifiers = [
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Developers"
	],
	package_dir = {"": "src"},
	python_requires = ">=3.7"
)


###############################################################################
# Coscine Python Client
# Copyright (c) 2018-2021 RWTH Aachen University
# Contact: coscine@itc.rwth-aachen.de
# Git: https://git.rwth-aachen.de/coscine/docs/public/coscine-python-client
# Please direct bug reports, feature requests or questions at the URL above
# by opening an issue.
###############################################################################
# Coscine is an open source project at RWTH Aachen University for
# the management of research data.
# Visit https://coscine.de for more information.
###############################################################################

from .form import Form
from .exceptions import *

###############################################################################

KEYS = [{
		"name": {
			"de": "Projektname",
			"en": "Project Name"
		},
		"flags": Form.REQUIRED,
		"field": "ProjectName"
	},{
		"name": {
			"de": "Anzeigename",
			"en": "Display Name"
		},
		"flags": Form.REQUIRED,
		"field": "DisplayName"
	},{
		"name": {
			"de": "Projektbeschreibung",
			"en": "Project Description"
		},
		"flags": Form.REQUIRED,
		"field": "Description"
	},{
		"name": {
			"de": "Principal Investigators",
			"en": "Principal Investigators"
		},
		"flags": Form.REQUIRED,
		"field": "PrincipleInvestigators"
	},{
		"name": {
			"de": "Projektstart",
			"en": "Project Start"
		},
		"flags": Form.REQUIRED,
		"field": "StartDate"
	},
	{
		"name": {
			"de": "Projektende",
			"en": "Project End"
		},
		"flags": Form.REQUIRED,
		"field": "EndDate"
	},{
		"name": {
			"de": "Disziplin",
			"en": "Discipline"
		},
		"flags": Form.REQUIRED | Form.CONTROLLED | Form.LIST,
		"field": "Discipline"
	},{
		"name": {
			"de": "Teilnehmende Organisation",
			"en": "Participating Organizations"
		},
		"flags": Form.REQUIRED | Form.CONTROLLED | Form.LIST,
		"field": "Organization"
	},{
		"name": {
			"de": "Projektschlagwörter",
			"en": "Project Keywords"
		},
		"flags": Form.NONE,
		"field": "Keywords"
	},{
		"name": {
			"de": "Grant ID",
			"en": "Grant ID"
		},
		"flags": Form.NONE,
		"field": "GrantId"
	},{
		"name": {
			"de": "Features",
			"en": "Features"
		},
		"flags": Form.CONTROLLED,
		"field": "Features"
}]

MAP = {
	"description": "Description",
	"displayName": "DisplayName",
	"startDate": "StartDate",
	"endDate": "EndDate",
	"keywords": "Keywords",
	"projectName": "ProjectName",
	"principleInvestigators": "PrincipleInvestigators",
	"grantId": "GrantId",
	"disciplines": "Discipline",
	"organizations": "Organization"
}

###############################################################################

class ProjectForm(Form):

	def __init__(self, lang, vocabulary, parent):
		super().__init__("Project Form", lang, KEYS, vocabulary)
		self.parent = parent

###############################################################################

	def parse(self, data):

		def _parse_disciplines(self, data):
			LSTR = {"en": "displayNameEn", "de": "displayNameDe"}[self._lang]
			disciplines = []
			for discipline in data:
				disciplines.append(discipline[LSTR])
			return disciplines

		def _parse_organizations(self, data):
			organizations = []
			for organization in data:
				vocabulary = self._vocabulary["Organization"]
				for entry in vocabulary.values():
					if entry["url"] == organization["url"]:
						organizations.append(entry["displayName"])
			return organizations

		for key in data:
			if key in MAP and data[key] != "":
				entry = self.entry(MAP[key])
				if entry is None:
					continue
				name = entry["name"][self._lang]
				flags = entry["flags"]
				value = data[key]
				if flags & Form.CONTROLLED or flags & Form.LIST:
					if key == "disciplines":
						self[name] = _parse_disciplines(self, value)
					elif key == "organizations":
						self[name] = _parse_organizations(self, value)
					elif key == "visibility":
						self[name] = value["displayName"]
				else:
					self[name] = value

###############################################################################

	def generate(self):

		def final_value(key, value):
			if self.is_controlled(key):
				return self._vocabulary[field][value]
			else:
				return value

		data = {}
		missing = []
		for key in self._keys:
			entry = self._keys[key]
			field = entry["field"]
			if key not in self.store:
				if self.is_required(key):
					missing.append(key)
			else:
				value = self.store[key]
				if entry["flags"] & ProjectForm.LIST:
					if type(value) is not list and type(value) is not tuple:
						raise ValueError("Expected iterable value for key %s" % key)
					data[field] = []
					for v in value:
						data[field].append(final_value(key, v))
				else:
					data[field] = final_value(key, value)

		if missing:
			raise RequirementError(missing)
		
		if self.parent:
			data["ParentId"] = self.parent.id

		data["Visibility"] = {
			"displayName": "Project Members",
			"id": "8ab9c883-eb0d-4402-aaad-2e4007badce6"
		}

		return data

###############################################################################